from enum import Enum
from typing import Any, Callable, List, Optional

from cg.constants.constants import SampleType
from cg.store.models import Customer, Sample
from sqlalchemy import or_
from sqlalchemy.orm import Query


def filter_samples_by_internal_id(internal_id: str, samples: Query, **kwargs) -> Query:
    """Return sample by internal id."""
    return samples.filter(Sample.internal_id == internal_id)


def filter_samples_by_name(name: str, samples: Query, **kwargs) -> Query:
    """Return sample with sample name."""
    return samples.filter(Sample.name == name)


def filter_samples_with_type(samples: Query, tissue_type: SampleType, **kwargs) -> Query:
    """Return samples with sample type."""
    is_tumour: bool = tissue_type == SampleType.TUMOR
    return samples.filter(Sample.is_tumour == is_tumour)


def filter_samples_with_loqusdb_id(samples: Query, **kwargs) -> Query:
    """Return samples with a loqusdb ID."""
    return samples.filter(Sample.loqusdb_id.isnot(None))


def filter_samples_without_loqusdb_id(samples: Query, **kwargs) -> Query:
    """Return samples without a loqusdb ID."""
    return samples.filter(Sample.loqusdb_id.is_(None))


def filter_samples_by_entry_id(entry_id: int, samples: Query, **kwargs) -> Query:
    """Return sample with entry id."""
    return samples.filter_by(id=entry_id)


def filter_samples_is_delivered(samples: Query, **kwargs) -> Query:
    """Return delivered samples."""
    return samples.filter(Sample.delivered_at.isnot(None))


def filter_samples_is_not_delivered(samples: Query, **kwargs) -> Query:
    """Return samples that are not delivered."""
    return samples.filter(Sample.delivered_at.is_(None))


def filter_samples_by_invoice_id(samples: Query, invoice_id: int, **kwargs) -> Query:
    """Return samples by invoice_id"""
    return samples.filter(Sample.invoice_id == invoice_id)


def filter_samples_without_invoice_id(samples: Query, **kwargs) -> Query:
    """Return samples that are not attached to an invoice."""
    return samples.filter(Sample.invoice_id.is_(None))


def filter_samples_is_not_down_sampled(samples: Query, **kwargs) -> Query:
    """Return samples that are not down sampled."""
    return samples.filter(Sample.downsampled_to.is_(None))


def filter_samples_is_sequenced(samples: Query, **kwargs) -> Query:
    """Return samples that are sequenced."""
    return samples.filter(Sample.reads_updated_at.isnot(None))


def filter_samples_is_not_sequenced(samples: Query, **kwargs) -> Query:
    """Return samples that are not sequenced."""
    return samples.filter(Sample.reads_updated_at.is_(None))


def filter_samples_do_invoice(samples: Query, **kwargs) -> Query:
    """Return samples that should be invoiced."""
    return samples.filter(Sample.no_invoice.is_(False))


def filter_samples_by_entry_customer_ids(
    samples: Query, customer_entry_ids: List[int], **kwargs
) -> Query:
    """Return samples by customer id."""
    return samples.filter(Sample.customer_id.in_(customer_entry_ids))


def filter_samples_is_received(samples: Query, **kwargs) -> Query:
    """Return samples that are received."""
    return samples.filter(Sample.received_at.isnot(None))


def filter_samples_is_not_received(samples: Query, **kwargs) -> Query:
    """Return samples that are not received."""
    return samples.filter(Sample.received_at.is_(None))


def filter_samples_is_prepared(samples: Query, **kwargs) -> Query:
    """Return samples that are prepared."""
    return samples.filter(Sample.prepared_at.isnot(None))


def filter_samples_is_not_prepared(samples: Query, **kwargs) -> Query:
    """Return samples that are not prepared."""
    return samples.filter(Sample.prepared_at.is_(None))


def filter_samples_by_subject_id(samples: Query, subject_id: str, **kwargs) -> Query:
    """Return samples by subject id."""
    return samples.filter(Sample.subject_id == subject_id)


def filter_samples_is_tumour(samples: Query, **kwargs) -> Query:
    """Return samples that are tumour."""
    return samples.filter(Sample.is_tumour.is_(True))


def filter_samples_is_not_tumour(samples: Query, **kwargs) -> Query:
    """Return samples that are not tumour."""
    return samples.filter(Sample.is_tumour.is_(False))


def filter_samples_by_internal_id_pattern(
    samples: Query, internal_id_pattern: str, **kwargs
) -> Query:
    """Return samples matching the internal id pattern."""
    return samples.filter(Sample.internal_id.like(f"%{internal_id_pattern}%"))


def filter_samples_by_internal_id_or_name_search(
    samples: Query, search_pattern: str, **kwargs
) -> Query:
    """Return samples matching the internal id or name search."""
    return samples.filter(
        or_(
            Sample.name.like(f"%{search_pattern}%"),
            Sample.internal_id.like(f"%{search_pattern}%"),
        )
    )


def filter_samples_by_customer(samples: Query, customer: Customer, **kwargs) -> Query:
    """Return samples by customer."""
    return samples.filter(Sample.customer == customer)


def order_samples_by_created_at_desc(samples: Query, **kwargs) -> Query:
    """Return samples ordered by created_at descending."""
    return samples.order_by(Sample.created_at.desc())


def filter_samples_by_identifier_name_and_value(
    samples: Query, identifier_name: str, identifier_value: Any, **kwargs
) -> Query:
    """Filters the sample query by the given identifier name and value."""
    return samples.filter(getattr(Sample, identifier_name) == identifier_value)


def apply_sample_filter(
    filter_functions: List[Callable],
    samples: Query,
    entry_id: Optional[int] = None,
    internal_id: Optional[str] = None,
    tissue_type: Optional[SampleType] = None,
    data_analysis: Optional[str] = None,
    invoice_id: Optional[int] = None,
    customer_entry_ids: Optional[List[int]] = None,
    subject_id: Optional[str] = None,
    name: Optional[str] = None,
    customer: Optional[Customer] = None,
    name_pattern: Optional[str] = None,
    internal_id_pattern: Optional[str] = None,
    search_pattern: Optional[str] = None,
    identifier_name: str = None,
    identifier_value: Any = None,
) -> Query:
    """Apply filtering functions to the sample queries and return filtered results."""

    for filter_function in filter_functions:
        samples: Query = filter_function(
            samples=samples,
            entry_id=entry_id,
            internal_id=internal_id,
            tissue_type=tissue_type,
            data_analysis=data_analysis,
            invoice_id=invoice_id,
            customer_entry_ids=customer_entry_ids,
            subject_id=subject_id,
            name=name,
            customer=customer,
            name_pattern=name_pattern,
            internal_id_pattern=internal_id_pattern,
            search_pattern=search_pattern,
            identifier_name=identifier_name,
            identifier_value=identifier_value,
        )
    return samples


class SampleFilter(Enum):
    """Define Sample filter functions."""

    FILTER_BY_CUSTOMER: Callable = filter_samples_by_customer
    FILTER_BY_CUSTOMER_ENTRY_IDS: Callable = filter_samples_by_entry_customer_ids
    FILTER_BY_ENTRY_ID: Callable = filter_samples_by_entry_id
    FILTER_BY_IDENTIFIER_NAME_AND_VALUE: Callable = filter_samples_by_identifier_name_and_value
    FILTER_BY_INTERNAL_ID: Callable = filter_samples_by_internal_id
    FILTER_BY_INTERNAL_ID_OR_NAME_SEARCH: Callable = filter_samples_by_internal_id_or_name_search
    FILTER_BY_INTERNAL_ID_PATTERN: Callable = filter_samples_by_internal_id_pattern
    FILTER_BY_INVOICE_ID: Callable = filter_samples_by_invoice_id
    FILTER_BY_SAMPLE_NAME: Callable = filter_samples_by_name
    FILTER_BY_SUBJECT_ID: Callable = filter_samples_by_subject_id
    FILTER_DO_INVOICE: Callable = filter_samples_do_invoice
    FILTER_HAS_NO_INVOICE_ID: Callable = filter_samples_without_invoice_id
    FILTER_IS_DELIVERED: Callable = filter_samples_is_delivered
    FILTER_IS_NOT_DELIVERED: Callable = filter_samples_is_not_delivered
    FILTER_IS_NOT_DOWN_SAMPLED: Callable = filter_samples_is_not_down_sampled
    FILTER_IS_PREPARED: Callable = filter_samples_is_prepared
    FILTER_IS_NOT_PREPARED: Callable = filter_samples_is_not_prepared
    FILTER_IS_RECEIVED: Callable = filter_samples_is_received
    FILTER_IS_NOT_RECEIVED: Callable = filter_samples_is_not_received
    FILTER_IS_SEQUENCED: Callable = filter_samples_is_sequenced
    FILTER_IS_NOT_SEQUENCED: Callable = filter_samples_is_not_sequenced
    FILTER_IS_TUMOUR: Callable = filter_samples_is_tumour
    FILTER_IS_NOT_TUMOUR: Callable = filter_samples_is_not_tumour
    FILTER_WITH_LOQUSDB_ID: Callable = filter_samples_with_loqusdb_id
    FILTER_WITHOUT_LOQUSDB_ID: Callable = filter_samples_without_loqusdb_id
    FILTER_WITH_TYPE: Callable = filter_samples_with_type
    ORDER_BY_CREATED_AT_DESC: Callable = order_samples_by_created_at_desc



"""Module for Flask-Admin views"""
from datetime import datetime
from gettext import gettext
from typing import List, Union

from cgmodels.cg.constants import Pipeline
from flask import flash, redirect, request, session, url_for
from flask_admin.actions import action
from flask_admin.contrib.sqla import ModelView
from flask_dance.contrib.google import google
from markupsafe import Markup

from cg.constants.constants import CaseActions, DataDelivery
from cg.server.ext import db
from cg.store.models import Sample
from cg.utils.flask.enum import SelectEnumField


class BaseView(ModelView):
    """Base for the specific views."""

    def is_accessible(self):
        user = db.get_user_by_email(email=session.get("user_email"))
        return bool(google.authorized and user and user.is_admin)

    def inaccessible_callback(self, name, **kwargs):
        # redirect to login page if user doesn't have access
        return redirect(url_for("google.login", next=request.url))


def view_priority(unused1, unused2, model, unused3):
    """column formatter for priority"""
    del unused1, unused2, unused3
    return Markup("%s" % model.priority.name) if model else ""


def view_family_sample_link(unused1, unused2, model, unused3):
    """column formatter to open the family-sample view"""

    del unused1, unused2, unused3

    return Markup(
        "<a href='%s'>%s</a>"
        % (url_for("familysample.index_view", search=model.internal_id), model.internal_id)
    )


def is_external_application(unused1, unused2, model, unused3):
    """column formatter to open this view"""
    del unused1, unused2, unused3
    return model.application_version.application.is_external if model.application_version else ""


class ApplicationView(BaseView):
    """Admin view for Model.Application"""

    column_editable_list = [
        "description",
        "is_accredited",
        "target_reads",
        "percent_reads_guaranteed",
        "comment",
        "prep_category",
        "sequencing_depth",
        "min_sequencing_depth",
        "is_external",
        "turnaround_time",
        "sample_concentration",
        "priority_processing",
        "is_archived",
    ]
    column_exclude_list = [
        "minimum_order",
        "sample_amount",
        "sample_volume",
        "details",
        "limitations",
        "created_at",
        "updated_at",
        "category",
    ]
    column_filters = ["prep_category", "is_accredited"]
    column_searchable_list = ["tag", "prep_category"]
    form_excluded_columns = ["category", "versions"]

    @staticmethod
    def view_application_link(unused1, unused2, model, unused3):
        """column formatter to open this view"""
        del unused1, unused2, unused3
        return (
            Markup(
                "<a href='%s'>%s</a>"
                % (
                    url_for("application.index_view", search=model.application.tag),
                    model.application.tag,
                )
            )
            if model.application
            else ""
        )


class ApplicationVersionView(BaseView):
    """Admin view for Model.ApplicationVersion"""

    column_default_sort = ("valid_from", True)
    column_editable_list = [
        "valid_from",
        "price_standard",
        "price_priority",
        "price_express",
        "price_clinical_trials",
        "price_research",
    ]
    column_exclude_list = ["created_at", "updated_at"]
    column_filters = ["version", "application.tag"]
    column_formatters = {"application": ApplicationView.view_application_link}
    column_searchable_list = ["application.tag"]
    edit_modal = True
    form_excluded_columns = ["samples", "pools", "microbial_samples"]


class BedView(BaseView):
    """Admin view for Model.Bed"""

    column_default_sort = "name"
    column_editable_list = ["comment"]
    column_exclude_list = ["created_at"]
    column_filters = []
    column_searchable_list = ["name"]
    form_excluded_columns = ["created_at", "updated_at"]

    @staticmethod
    def view_bed_link(unused1, unused2, model, unused3):
        """column formatter to open this view"""
        del unused1, unused2, unused3
        return (
            Markup(
                "<a href='%s'>%s</a>"
                % (url_for("bed.index_view", search=model.bed.name), model.bed.name)
            )
            if model.bed
            else ""
        )


class BedVersionView(BaseView):
    """Admin view for Model.BedVersion"""

    column_default_sort = ("updated_at", True)
    column_editable_list = ["shortname", "filename", "comment", "designer", "checksum"]
    column_exclude_list = ["created_at"]
    form_excluded_columns = ["created_at", "updated_at", "samples"]
    column_filters = []
    column_formatters = {"bed": BedView.view_bed_link}
    column_searchable_list = ["bed.name"]
    edit_modal = True


class CustomerView(BaseView):
    """Admin view for Model.Customer"""

    column_editable_list = [
        "collaborations",
        "comment",
        "delivery_contact",
        "loqus_upload",
        "primary_contact",
        "priority",
        "return_samples",
        "scout_access",
    ]
    column_list = [
        "comment",
        "delivery_contact",
        "internal_id",
        "name",
        "primary_contact",
        "priority",
        "project_account_KI",
        "project_account_kth",
        "return_samples",
        "scout_access",
    ]
    column_filters = ["priority", "scout_access"]
    column_searchable_list = ["internal_id", "name"]
    form_excluded_columns = ["families", "samples", "pools", "orders", "invoices"]


class CollaborationView(BaseView):
    """Admin view for Model.CustomerGroup"""

    column_editable_list = ["name"]
    column_filters = []
    column_hide_backrefs = False
    column_list = ("internal_id", "name", "customers")
    column_searchable_list = ["internal_id", "name"]


class FamilyView(BaseView):
    """Admin view for Model.Family"""

    column_default_sort = ("created_at", True)
    column_editable_list = ["action", "comment"]
    column_exclude_list = ["created_at", "_cohorts", "synopsis"]
    column_filters = [
        "customer.internal_id",
        "priority",
        "action",
        "data_analysis",
        "data_delivery",
        "tickets",
    ]
    column_formatters = {
        "internal_id": view_family_sample_link,
        "priority": view_priority,
    }
    column_searchable_list = [
        "internal_id",
        "name",
        "customer.internal_id",
        "tickets",
    ]
    form_excluded_columns = [
        "analyses",
        "_cohorts",
        "links",
        "synopsis",
    ]
    form_extra_fields = {
        "data_analysis": SelectEnumField(enum_class=Pipeline),
        "data_delivery": SelectEnumField(enum_class=DataDelivery),
    }

    @staticmethod
    def view_family_link(unused1, unused2, model, unused3):
        """column formatter to open this view"""
        del unused1, unused2, unused3
        markup = ""
        if model.family:
            markup += Markup(
                " <a href='%s'>%s</a>"
                % (url_for("family.index_view", search=model.family.internal_id), model.family)
            )

        return markup

    @action(
        "set_hold",
        "Set action to hold",
        "Are you sure you want to set the action for selected families to hold?",
    )
    def action_set_hold(self, ids: List[str]):
        self.set_action_for_cases(action=CaseActions.HOLD, case_entry_ids=ids)

    @action(
        "set_empty",
        "Set action to Empty",
        "Are you sure you want to set the action for selected families to Empty?",
    )
    def action_set_empty(self, ids: List[str]):
        self.set_action_for_cases(action=None, case_entry_ids=ids)

    def set_action_for_cases(self, action: Union[CaseActions, None], case_entry_ids: List[str]):
        try:
            for entry_id in case_entry_ids:
                family = db.get_case_by_entry_id(entry_id=entry_id)
                if family:
                    family.action = action

            db.session.commit()

            num_families = len(case_entry_ids)
            action_message = (
                f"Families were set to {action}."
                if num_families == 1
                else f"{num_families} families were set to {action}."
            )
            flash(action_message)

        except Exception as ex:
            if not self.handle_view_exception(ex):
                raise

            flash(gettext(f"Failed to set family action. {str(ex)}"))


class FlowcellView(BaseView):
    """Admin view for Model.Flowcell"""

    column_default_sort = ("sequenced_at", True)
    column_editable_list = ["status"]
    column_exclude_list = ["archived_at"]
    column_filters = ["sequencer_type", "sequencer_name", "status"]
    column_searchable_list = ["name"]

    @staticmethod
    def view_flow_cell_link(unused1, unused2, model, unused3):
        """column formatter to open this view"""
        del unused1, unused2, unused3
        return Markup(
            "<a href='%s'>%s</a>"
            % (url_for("flowcell.index_view", search=model.flowcell.name), model.flowcell.name)
        )


class InvoiceView(BaseView):
    """Admin view for Model.Invoice"""

    column_default_sort = ("created_at", True)
    column_editable_list = ["comment"]
    column_list = (
        "id",
        "customer",
        "created_at",
        "updated_at",
        "invoiced_at",
        "comment",
        "discount",
        "price",
    )
    column_searchable_list = ["customer.internal_id", "customer.name", "id"]

    @staticmethod
    def view_invoice_link(unused1, unused2, model, unused3):
        """column formatter to open this view"""
        del unused1, unused2, unused3
        return (
            Markup(
                "<a href='%s'>%s</a>"
                % (
                    url_for("invoice.index_view", search=model.invoice.id),
                    model.invoice.invoiced_at.date()
                    if model.invoice.invoiced_at
                    else "In progress",
                )
            )
            if model.invoice
            else ""
        )


class AnalysisView(BaseView):
    """Admin view for Model.Analysis"""

    column_default_sort = ("created_at", True)
    column_editable_list = ["is_primary"]
    column_filters = ["pipeline", "pipeline_version", "is_primary"]
    column_formatters = {"family": FamilyView.view_family_link}
    column_searchable_list = [
        "family.internal_id",
        "family.name",
    ]
    form_extra_fields = {"pipeline": SelectEnumField(enum_class=Pipeline)}


class OrganismView(BaseView):
    """Admin view for Model.Organism"""

    column_default_sort = ("created_at", True)
    column_editable_list = ["internal_id", "name", "reference_genome", "comment"]
    column_searchable_list = ["internal_id", "name", "reference_genome"]


class PanelView(BaseView):
    """Admin view for Model.Panel"""

    column_editable_list = ["current_version", "name"]
    column_filters = ["customer.internal_id"]
    column_searchable_list = ["customer.internal_id", "name", "abbrev"]
    create_modal = True
    edit_modal = True


class PoolView(BaseView):
    """Admin view for Model.Pool"""

    column_default_sort = ("created_at", True)
    column_editable_list = ["ticket"]
    column_filters = ["customer.internal_id", "application_version.application"]
    column_formatters = {"invoice": InvoiceView.view_invoice_link}
    column_searchable_list = ["name", "order", "ticket", "customer.internal_id"]


class SampleView(BaseView):
    """Admin view for Model.Sample"""

    column_exclude_list = [
        "age_at_sampling",
        "invoiced_at",
        "_phenotype_groups",
        "_phenotype_terms",
    ]
    column_default_sort = ("created_at", True)
    column_editable_list = [
        "comment",
        "downsampled_to",
        "is_tumour",
        "reads_updated_at",
        "sex",
    ]
    column_filters = ["customer.internal_id", "priority", "sex", "application_version.application"]
    column_formatters = {
        "is_external": is_external_application,
        "internal_id": view_family_sample_link,
        "invoice": InvoiceView.view_invoice_link,
        "priority": view_priority,
    }
    column_searchable_list = [
        "internal_id",
        "name",
        "subject_id",
        "customer.internal_id",
        "original_ticket",
    ]
    form_excluded_columns = [
        "age_at_sampling",
        "deliveries",
        "father_links",
        "flowcells",
        "invoiced_at",
        "invoice",
        "is_external",
        "_phenotype_groups",
        "_phenotype_terms",
        "links",
        "mother_links",
    ]

    @staticmethod
    def view_sample_link(unused1, unused2, model, unused3):
        """column formatter to open this view"""
        del unused1, unused2, unused3
        return (
            Markup(
                "<a href='%s'>%s</a>"
                % (url_for("sample.index_view", search=model.sample.internal_id), model.sample)
            )
            if model.sample
            else ""
        )

    @action(
        "cancel_samples",
        "Cancel samples",
        "Are you sure you want to cancel the selected samples?",
    )
    def cancel_samples(self, entry_ids: List[str]) -> None:
        """
        Action for cancelling samples:
            - Comments each sample being cancelled with date and user.
            - Deletes any relationship between the cancelled samples and cases.
            - Deletes any cases that only contain samples being cancelled.
        """
        all_associated_case_ids: set = set()

        for entry_id in entry_ids:
            sample: Sample = db.get_sample_by_entry_id(entry_id=int(entry_id))

            sample_case_ids: List[str] = [
                case_sample.family.internal_id for case_sample in sample.links
            ]
            all_associated_case_ids.update(sample_case_ids)

            db.delete_relationships_sample(sample=sample)
            self.write_cancel_comment(sample=sample)

        case_ids: List[str] = list(all_associated_case_ids)
        db.delete_cases_without_samples(case_internal_ids=case_ids)
        cases_with_remaining_samples: List[str] = db.filter_cases_with_samples(case_ids=case_ids)

        self.display_cancel_confirmation(
            sample_entry_ids=entry_ids, remaining_cases=cases_with_remaining_samples
        )

    def write_cancel_comment(self, sample: Sample) -> None:
        """Add comment to sample with date and user cancelling the sample."""
        user_name: str = db.get_user_by_email(session.get("user_email")).name
        date: str = datetime.now().strftime("%Y-%m-%d")
        comment: str = f"Cancelled {date} by {user_name}"

        db.add_sample_comment(sample=sample, comment=comment)

    def display_cancel_confirmation(
        self, sample_entry_ids: List[str], remaining_cases: List[str]
    ) -> None:
        """Show a summary of the cancelled samples and any cases in which other samples were present."""
        samples: str = "sample" if len(sample_entry_ids) == 1 else "samples"
        cases: str = "case" if len(remaining_cases) == 1 else "cases"

        message: str = f"Cancelled {len(sample_entry_ids)} {samples}. "
        case_message: str = ""

        for case_id in remaining_cases:
            case_message = f"{case_message} {case_id},"

        case_message = case_message.strip(",")

        if remaining_cases:
            message += (
                f"Found {len(remaining_cases)} {cases} with additional samples: {case_message}."
            )
        else:
            message += "No case contained additional samples."

        flash(message=message)


class DeliveryView(BaseView):
    """Admin view for Model.Delivery"""

    column_default_sort = ("id", True)
    column_filters = ["sample.internal_id"]
    column_formatters = {"sample": SampleView.view_sample_link}
    column_searchable_list = ["sample.internal_id"]
    create_modal = True
    edit_modal = True


class FamilySampleView(BaseView):
    """Admin view for Model.FamilySample"""

    column_default_sort = ("created_at", True)
    column_editable_list = ["status"]
    column_filters = ["status"]
    column_formatters = {
        "family": FamilyView.view_family_link,
        "sample": SampleView.view_sample_link,
    }
    column_searchable_list = ["family.internal_id", "family.name", "sample.internal_id"]
    create_modal = True
    edit_modal = True


class UserView(BaseView):
    """Admin view for Model.User"""

    column_default_sort = "name"
    column_editable_list = ["order_portal_login"]
    column_filters = ["is_admin", "order_portal_login", "customers"]
    column_hide_backrefs = False
    column_list = ("name", "email", "is_admin", "order_portal_login", "customers")
    column_searchable_list = ["name", "email"]
    create_modal = True
    edit_modal = True


class SampleLaneSequencingMetricsView(BaseView):
    """Admin view for the Model.SampleLaneSequencingMetrics."""

    column_filters = ["sample_internal_id", "flow_cell_name"]
    column_searchable_list = ["sample_internal_id", "flow_cell_name"]
    create_modal = True
    edit_modal = True

    column_formatters = {
        "flowcell": FlowcellView.view_flow_cell_link,
        "sample": SampleView.view_sample_link,
    }



from pathlib import Path

from cg.cli.demultiplex.copy_novaseqx_demultiplex_data import get_latest_analysis_path
from cg.cli.demultiplex.demux import hardlink_flow_cell_analysis_data, is_ready_for_post_processing
from cg.constants.demultiplexing import DemultiplexingDirsAndFiles
from cg.utils.files import get_all_files_in_dir


def test_flow_cell_is_ready_for_post_processing(
    novaseqx_flow_cell_dir_with_analysis_data: Path, demultiplexed_runs_flow_cell_directory: Path
):
    # GIVEN a flow cell which is ready for post processing

    # WHEN checking if the flow cell is ready for post processing
    is_flow_cell_ready: bool = is_ready_for_post_processing(
        novaseqx_flow_cell_dir_with_analysis_data, demultiplexed_runs_flow_cell_directory
    )

    # THEN the flow cell is ready for post processing
    assert is_flow_cell_ready


def test_flow_cell_is_not_ready_for_post_processing_without_analysis(
    novaseqx_flow_cell_analysis_incomplete: Path, demultiplexed_runs_flow_cell_directory: Path
):
    # GIVEN a flow cell for which analysis is not completed

    # WHEN checking if the flow cell is ready for post processing
    is_flow_cell_ready: bool = is_ready_for_post_processing(
        novaseqx_flow_cell_analysis_incomplete, demultiplexed_runs_flow_cell_directory
    )

    # THEN the flow cell is not ready for post processing
    assert not is_flow_cell_ready


def test_flow_cell_is_not_ready_for_post_processing_when_not_demultiplexed(
    demultiplex_not_complete_novaseqx_flow_cell: Path, demultiplexed_runs_flow_cell_directory: Path
):
    # GIVEN a flow cell for which demultiplexing is not completed

    # WHEN checking if the flow cell is ready for post processing
    is_flow_cell_ready: bool = is_ready_for_post_processing(
        demultiplex_not_complete_novaseqx_flow_cell, demultiplexed_runs_flow_cell_directory
    )

    # THEN the flow cell is not ready for post processing
    assert not is_flow_cell_ready


def test_flow_cell_is_not_ready_when_already_post_processed(
    post_processed_novaseqx_flow_cell: Path, demultiplexed_runs_flow_cell_directory: Path
):
    # GIVEN a flow cell for which post processing is done

    # WHEN checking if the flow cell is ready for post processing
    is_flow_cell_ready: bool = is_ready_for_post_processing(
        post_processed_novaseqx_flow_cell, demultiplexed_runs_flow_cell_directory
    )

    # THEN the flow cell is not ready for post processing as the the flow cell has already been post processed
    assert not is_flow_cell_ready


def test_previously_copied_flow_cell_is_not_ready(
    novaseqx_flow_cell_dir_with_analysis_data: Path, demultiplexed_runs_flow_cell_directory: Path
):
    # GIVEN a flow cell which already exists in demultiplexed runs
    Path(
        demultiplexed_runs_flow_cell_directory, novaseqx_flow_cell_dir_with_analysis_data.name
    ).mkdir()

    # WHEN checking if the flow cell is ready for post processing
    is_flow_cell_ready: bool = is_ready_for_post_processing(
        novaseqx_flow_cell_dir_with_analysis_data, demultiplexed_runs_flow_cell_directory
    )

    # THEN the flow cell is not ready for post processing
    assert not is_flow_cell_ready


def test_get_latest_analysis_version_path(
    novaseqx_flow_cell_dir_with_analysis_data: Path,
    novaseqx_latest_analysis_version: str,
):
    # GIVEN a flow cell which is ready to be post processed

    # WHEN extracting the latest analysis version path
    analysis_path: Path = get_latest_analysis_path(novaseqx_flow_cell_dir_with_analysis_data)

    # THEN the latest analysis version path is returned
    latest_analysis_path: Path = Path(
        novaseqx_flow_cell_dir_with_analysis_data,
        DemultiplexingDirsAndFiles.ANALYSIS,
        novaseqx_latest_analysis_version,
    )
    assert analysis_path == latest_analysis_path


def test_copy_novaseqx_flow_cell(
    demultiplexed_runs_flow_cell_directory: Path,
    novaseqx_flow_cell_dir_with_analysis_data: Path,
    novaseq_x_flow_cell_full_name: str,
):
    # GIVEN a demultiplexed runs directory for a NovaseqX flow cell
    demultiplexed_runs_novaseqx_dir: Path = Path(
        demultiplexed_runs_flow_cell_directory, novaseq_x_flow_cell_full_name
    )
    # WHEN copying the flow cell analysis data to demultiplexed runs NovaseqX directory
    hardlink_flow_cell_analysis_data(
        novaseqx_flow_cell_dir_with_analysis_data, demultiplexed_runs_flow_cell_directory
    )

    # THEN the data contains everything from the analysis folder
    analysis: Path = get_latest_analysis_path(novaseqx_flow_cell_dir_with_analysis_data)
    analysis_data_path: Path = Path(analysis / DemultiplexingDirsAndFiles.DATA)

    original_files = get_all_files_in_dir(analysis_data_path)
    copied_files = get_all_files_in_dir(demultiplexed_runs_novaseqx_dir)

    # THEN the original files are the same as the copied files
    assert original_files == copied_files



from typing import List
from pathlib import Path


from cg.apps.sequencing_metrics_parser.parsers.bcl2fastq_to_sequencing_statistics import (
    create_bcl2fastq_undetermined_metrics,
)
from cg.store.models import SampleLaneSequencingMetrics


def test_create_sequencing_statistics_from_bcl2fastq_flow_cell(bcl2fastq_flow_cell_path: Path):
    """Test creating metrics for undetermined reads on a flow cell demultiplexed with bcl2fastq."""
    # GIVEN a flow cell demultiplexed with bcl2fastq with one lane, one sample and two undetermined reads

    # WHEN creating undetermined metrics for the lane and sample
    metrics: List[SampleLaneSequencingMetrics] = create_bcl2fastq_undetermined_metrics(
        bcl2fastq_flow_cell_path=bcl2fastq_flow_cell_path,
        non_pooled_lane_sample_pairs=[(1, "sample_id")],
    )

    # THEN a list of metrics is returned
    assert isinstance(metrics, list)
    assert all(isinstance(item, SampleLaneSequencingMetrics) for item in metrics)
    assert len(metrics) == 1

    # THEN the undetermined metrics have been assigned the correct sample id and lane
    assert metrics[0].sample_internal_id == "sample_id"
    assert metrics[0].flow_cell_lane_number == 1

    # THEN the two undetermined paired reads in each tile were combined
    assert metrics[0].sample_total_reads_in_lane == 4


def test_create_undetermined_metrics_for_invalid_lane(bcl2fastq_flow_cell_path: Path):
    """Test creating metrics for undetrmined reads on a flow cell demultiplexed with bcl2fastq with an invalid lane."""
    # GIVEN a flow cell demultiplexed with bcl2fastq

    # WHEN creating metrics for a non existing lane on the flow cell
    metrics: List[SampleLaneSequencingMetrics] = create_bcl2fastq_undetermined_metrics(
        bcl2fastq_flow_cell_path=bcl2fastq_flow_cell_path,
        non_pooled_lane_sample_pairs=[(2, "sample_id")],
    )

    # THEN no metrics are returned
    assert not metrics



"""Tests cli methods to create the case config for balsamic"""

import logging
from pathlib import Path

from _pytest.logging import LogCaptureFixture

from cg.cli.workflow.balsamic.base import config_case
from click.testing import CliRunner

from cg.models.cg_config import CGConfig

EXIT_SUCCESS = 0


def test_without_options(cli_runner: CliRunner, balsamic_context: CGConfig):
    """Test command without case_id."""
    # GIVEN NO case_id
    # WHEN dry running without anything specified
    result = cli_runner.invoke(config_case, obj=balsamic_context)
    # THEN command should mention argument
    assert result.exit_code != EXIT_SUCCESS
    assert "Missing argument" in result.output


def test_with_missing_case(
    cli_runner: CliRunner, balsamic_context: CGConfig, caplog: LogCaptureFixture
):
    """Test command with invalid case to start with."""
    caplog.set_level(logging.ERROR)
    # GIVEN case_id not in database
    case_id = "soberelephant"
    assert not balsamic_context.status_db.get_case_by_internal_id(internal_id=case_id)
    # WHEN running
    result = cli_runner.invoke(config_case, [case_id], obj=balsamic_context)
    # THEN command should NOT successfully call the command it creates
    assert result.exit_code != EXIT_SUCCESS
    # THEN ERROR log should be printed containing invalid case_id
    assert "could not be found in Status DB!" in caplog.text


def test_without_samples(
    cli_runner: CliRunner, balsamic_context: CGConfig, caplog: LogCaptureFixture
):
    """Test command with case_id and no samples."""
    caplog.set_level(logging.ERROR)
    # GIVEN case-id
    case_id = "no_sample_case"
    # WHEN dry running with dry specified
    result = cli_runner.invoke(config_case, [case_id, "--dry-run"], obj=balsamic_context)
    # THEN command should print the balsamic command-string
    assert result.exit_code != EXIT_SUCCESS
    # THEN warning should be printed that no config file is found
    assert case_id in caplog.text
    assert "has no samples" in caplog.text


def test_dry(cli_runner: CliRunner, balsamic_context: CGConfig, caplog: LogCaptureFixture):
    """Test command with --dry option."""
    caplog.set_level(logging.INFO)
    # GIVEN a VALID case_id
    case_id = "balsamic_case_wgs_paired"
    # WHEN dry running with dry option specified
    result = cli_runner.invoke(config_case, [case_id, "--dry-run"], obj=balsamic_context)
    # THEN command should print the balsamic command string
    assert result.exit_code == EXIT_SUCCESS
    # THEN gry run, balsamic and case_id should be in log
    assert "Dry run" in caplog.text
    assert "balsamic" in caplog.text
    assert case_id in caplog.text


def test_genome_version(
    cli_runner: CliRunner, balsamic_context: CGConfig, caplog: LogCaptureFixture
):
    """Test command with --genome-version option."""
    caplog.set_level(logging.INFO)
    # GIVEN a VALID case_id and genome_version
    case_id = "balsamic_case_wgs_paired"
    option_key = "--genome-version"
    option_value = "canfam3"
    # WHEN dry running with genome option specified
    result = cli_runner.invoke(
        config_case,
        [case_id, "--dry-run", option_key, option_value],
        obj=balsamic_context,
    )
    # THEN command should be generated successfully
    assert result.exit_code == EXIT_SUCCESS
    # THEN dry-print should include the option key and value
    assert option_key in caplog.text
    assert option_value in caplog.text


def test_target_bed(
    cli_runner: CliRunner,
    balsamic_context: CGConfig,
    balsamic_bed_2_path: Path,
    caplog: LogCaptureFixture,
):
    """Test command with --panel-bed option."""
    caplog.set_level(logging.INFO)
    # GIVEN VALID case_id of application type that requires BED
    case_id = "balsamic_case_tgs_single"
    option_key = "--panel-bed"
    option_value = balsamic_bed_2_path
    # WHEN dry running with PANEL BED option specified
    result = cli_runner.invoke(
        config_case,
        [case_id, "--dry-run", option_key, option_value],
        obj=balsamic_context,
    )
    # THEN command should be generated successfully
    assert result.exit_code == EXIT_SUCCESS
    # THEN dry-print should include the the option key and value
    assert option_key in caplog.text
    assert option_value in caplog.text


def test_target_bed_from_lims(
    cli_runner: CliRunner, balsamic_context: CGConfig, caplog: LogCaptureFixture
):
    """Test command without --target-bed option when BED can be retrieved from LIMS."""
    caplog.set_level(logging.INFO)
    # GIVEN case that bed-version set in lims with same version existing in status db
    case_id = "balsamic_case_tgs_single"

    # WHEN dry running
    result = cli_runner.invoke(config_case, [case_id, "--dry-run"], obj=balsamic_context)

    # THEN command should be generated successfully
    assert result.exit_code == EXIT_SUCCESS

    # THEN dry-print should include the bed_key and the bed_value including path
    assert "--panel-bed" in caplog.text
    assert ".bed" in caplog.text


def test_paired_wgs(balsamic_context: CGConfig, cli_runner: CliRunner, caplog: LogCaptureFixture):
    """Test with case_id that requires PAIRED WGS analysis."""
    caplog.set_level(logging.INFO)
    # GIVEN case_id containing ONE tumor, ONE normal, WGS application
    case_id = "balsamic_case_wgs_paired"
    # WHEN dry running
    result = cli_runner.invoke(config_case, [case_id, "--dry-run"], obj=balsamic_context)
    # THEN command is generated successfully
    assert result.exit_code == EXIT_SUCCESS
    # THEN dry-print should not include panel bed
    assert "--panel-bed" not in caplog.text
    # THEN tumor and normal options should be included in command
    assert "--tumor" in caplog.text
    assert "--normal" in caplog.text


def test_paired_panel(balsamic_context: CGConfig, cli_runner: CliRunner, caplog: LogCaptureFixture):
    """Test with case_id that requires PAIRED TGS analysis."""
    caplog.set_level(logging.INFO)
    # GIVEN case_id containing ONE tumor, ONE normal, TGS application
    case_id = "balsamic_case_tgs_paired"
    # WHEN dry running
    result = cli_runner.invoke(config_case, [case_id, "--dry-run"], obj=balsamic_context)
    # THEN command is generated successfully
    assert result.exit_code == EXIT_SUCCESS
    # THEN dry-print should include panel bed
    assert "--panel-bed" in caplog.text
    # THEN tumor and normal options should be included in command
    assert "--tumor" in caplog.text
    assert "--normal" in caplog.text


def test_pon_cnn(
    balsamic_context: CGConfig,
    cli_runner: CliRunner,
    balsamic_bed_1_path: str,
    balsamic_pon_1_path: str,
    caplog: LogCaptureFixture,
):
    """Test command with --pon-cnn option."""
    caplog.set_level(logging.INFO)
    # GIVEN VALID case_id of application with BED and PoN files
    case_id = "balsamic_case_tgs_paired"
    panel_key = "--panel-bed"
    panel_value = balsamic_bed_1_path
    pon_key = "--pon-cnn"
    pon_value = balsamic_pon_1_path
    # WHEN dry running with PANEL BED and PON CNN options specified
    result = cli_runner.invoke(
        config_case,
        [case_id, "--dry-run", panel_key, panel_value, pon_key, pon_value],
        obj=balsamic_context,
    )
    # THEN command should be generated successfully
    assert result.exit_code == EXIT_SUCCESS
    # THEN dry-print should include the PoN option key and value
    assert pon_key in caplog.text
    assert pon_value in caplog.text


def test_single_wgs(balsamic_context: CGConfig, cli_runner: CliRunner, caplog: LogCaptureFixture):
    """Test with case_id that requires SINGLE WGS analysis."""
    caplog.set_level(logging.INFO)
    # GIVEN case_id containing ONE tumor, WGS application
    case_id = "balsamic_case_wgs_single"
    # WHEN dry running
    result = cli_runner.invoke(config_case, [case_id, "--dry-run"], obj=balsamic_context)
    # THEN command is generated successfully
    assert result.exit_code == EXIT_SUCCESS
    # THEN dry-print should NOT include panel bed
    assert "--panel-bed" not in caplog.text
    # THEN tumor option should be included in command
    assert "--tumor" in caplog.text
    # THEN normal option should NOT be included in command
    assert "--normal" not in caplog.text


def test_single_panel(balsamic_context: CGConfig, cli_runner: CliRunner, caplog: LogCaptureFixture):
    """Test with case_id that requires SINGLE TGS analysis."""
    caplog.set_level(logging.INFO)
    # GIVEN case_id containing ONE tumor, TGS application
    case_id = "balsamic_case_tgs_single"
    # WHEN dry running
    result = cli_runner.invoke(config_case, [case_id, "--dry-run"], obj=balsamic_context)
    # THEN command is generated successfully
    assert result.exit_code == EXIT_SUCCESS
    # THEN dry-print should include panel bed
    assert "--panel-bed" in caplog.text
    # THEN tumor option should be included in command
    assert "--tumor" in caplog.text
    # THEN normal option should NOT be included in command
    assert "--normal" not in caplog.text


def test_error_single_wgs_panel_arg(
    balsamic_context: CGConfig, cg_dir: Path, cli_runner: CliRunner, caplog: LogCaptureFixture
):
    """Test with case_id that requires SINGLE WGS analysis and --panel-bed argument."""
    caplog.set_level(logging.ERROR)
    # GIVEN case_id containing ONE tumor, WGS application and panel bed argument
    case_id = "balsamic_case_wgs_single"
    panel_bed = Path(cg_dir, "balsamic_bed_1.bed")
    # WHEN dry running
    result = cli_runner.invoke(
        config_case,
        [case_id, "--dry-run", "--panel-bed", panel_bed],
        obj=balsamic_context,
    )
    # THEN command is NOT generated successfully
    assert result.exit_code != EXIT_SUCCESS
    # THEN log warning not to set panel for WGS should be printed
    assert "Cannot set panel_bed for WGS sample" in caplog.text


def test_single_normal_only(
    balsamic_context: CGConfig, cli_runner: CliRunner, caplog: LogCaptureFixture
):
    """Test with case_id that has only one normal sample."""
    caplog.set_level(logging.WARNING)

    # GIVEN case_id containing one normal sample
    case_id = "balsamic_case_tgs_single_error"

    # WHEN dry running
    result = cli_runner.invoke(config_case, [case_id, "--dry-run"], obj=balsamic_context)

    # THEN command should be generated successfully
    assert result.exit_code == EXIT_SUCCESS

    # THEN a warning log should be printed specifying that a normal sample will be used as tumor for Balsamic analysis
    assert (
        f"Only a normal sample was found for case {case_id}. Balsamic analysis will treat it as a tumor sample."
        in caplog.text
    )


def test_error_two_tumor(
    balsamic_context: CGConfig, cli_runner: CliRunner, caplog: LogCaptureFixture
):
    """Test with case_id that requires WGS analysis but has TWO tumor ONE normal samples."""
    caplog.set_level(logging.ERROR)
    # GIVEN case_id containing TWO tumor, ONE normal, TGS application
    case_id = "balsamic_case_tgs_paired_error"
    # WHEN dry running
    result = cli_runner.invoke(config_case, [case_id, "--dry-run"], obj=balsamic_context)
    # THEN command is NOT generated successfully
    assert result.exit_code != EXIT_SUCCESS
    # THEN log warning should be printed
    assert f"Case {case_id} has an invalid number of samples" in caplog.text


def test_error_mixed_application(
    balsamic_context: CGConfig, cli_runner: CliRunner, caplog: LogCaptureFixture
):
    """Test with case_id that has ONE tumor ONE normal samples marked for WGS and TGS analysis."""
    caplog.set_level(logging.ERROR)
    case_id = "balsamic_case_mixed_paired_error"
    # WHEN dry running
    result = cli_runner.invoke(config_case, [case_id, "--dry-run"], obj=balsamic_context)
    # THEN command is NOT generated successfully
    assert result.exit_code != EXIT_SUCCESS
    # THEN log warning should be printed
    assert "Multiple application types found" in caplog.text


def test_error_not_balsamic_application(
    balsamic_context: CGConfig, cli_runner: CliRunner, caplog: LogCaptureFixture
):
    """Test with case_id that has PAIRED samples marked for WGS and MIC analysis."""
    caplog.set_level(logging.ERROR)
    case_id = "balsamic_case_mixed_wgs_mic_paired_error"
    # WHEN dry running
    result = cli_runner.invoke(config_case, [case_id, "--dry-run"], obj=balsamic_context)
    # THEN command is NOT generated successfully
    assert result.exit_code != EXIT_SUCCESS
    # THEN log warning should be printed
    assert "not compatible with BALSAMIC" in caplog.text


def test_error_mixed_panel_bed_resque(
    balsamic_context: CGConfig, cli_runner: CliRunner, caplog: LogCaptureFixture
):
    """
    Test with case_id marked for PAIRED TGS analysis but different BED files in LIMS
    and supplying --panel-bed option should prevent error.
    """
    caplog.set_level(logging.INFO)
    # GIVEN case_id with mixed panel_bed in LIMS and a panel bed argument
    case_id = "balsamic_case_mixed_bed_paired_error"
    panel_bed = "BalsamicBed1"
    # WHEN dry running
    result = cli_runner.invoke(
        config_case,
        [case_id, "--dry-run", "--panel-bed", panel_bed],
        obj=balsamic_context,
    )
    # THEN command is generated successfully
    assert result.exit_code == EXIT_SUCCESS
    # THEN panel bed should override LIMS data and be present in command
    assert "--panel-bed" in caplog.text
    assert panel_bed in caplog.text


def test_error_two_normal(
    balsamic_context: CGConfig, cli_runner: CliRunner, caplog: LogCaptureFixture
):
    """Test with case_id containing ONE tumor and TWO normal samples."""
    caplog.set_level(logging.ERROR)
    case_id = "balsamic_case_wgs_paired_two_normal_error"
    # WHEN dry running
    result = cli_runner.invoke(config_case, [case_id, "--dry-run"], obj=balsamic_context)
    # THEN command is NOT generated successfully
    assert result.exit_code != EXIT_SUCCESS
    # THEN log warning should be printed
    assert f"Case {case_id} has an invalid number of samples" in caplog.text


def test_error_wes_panel(
    balsamic_context: CGConfig, cli_runner: CliRunner, caplog: LogCaptureFixture
):
    """Test with case_id containing ONE tumor and TWO normal samples."""
    caplog.set_level(logging.ERROR)
    case_id = "balsamic_case_wes_panel_error"
    # WHEN dry running
    result = cli_runner.invoke(config_case, [case_id, "--dry-run"], obj=balsamic_context)
    # THEN command is NOT generated successfully
    assert result.exit_code != EXIT_SUCCESS
    # THEN log warning should be printed
    assert "requires a bed file" in caplog.text



"""Schemas for scout serialisation"""

from datetime import datetime
from typing import Dict, List, Optional

from cg.apps.scout.validators import set_gender_if_other, set_parent_if_missing
from cg.constants.gene_panel import GENOME_BUILD_37
from cg.constants.subject import Gender, PlinkGender, PlinkPhenotypeStatus
from pydantic import BaseModel, BeforeValidator, Field, Strict
from typing_extensions import Annotated, Literal


class Individual(BaseModel):
    bam_file: Optional[str] = None
    individual_id: str
    sex: Annotated[
        Literal[PlinkGender.UNKNOWN, PlinkGender.MALE, PlinkGender.FEMALE, Gender.OTHER],
        BeforeValidator(set_gender_if_other),
    ]
    father: Annotated[str, BeforeValidator(set_parent_if_missing)]
    mother: Annotated[str, BeforeValidator(set_parent_if_missing)]
    phenotype: PlinkPhenotypeStatus
    analysis_type: str = "wgs"


class Panel(BaseModel):
    panel_name: str


class Phenotype(BaseModel):
    phenotype_id: str
    feature: str


class Gene(BaseModel):
    hgnc_id: int
    hgnc_symbol: Optional[str] = None
    region_annotation: Optional[str] = None
    functional_annotation: Optional[str] = None
    sift_prediction: Optional[str] = None
    polyphen_prediction: Optional[str] = None


class DiagnosisPhenotypes(BaseModel):
    disease_nr: int
    disease_id: str
    description: str
    individuals: Optional[List[Dict[str, str]]] = None


class ScoutExportCase(BaseModel):
    id: str = Field(str, alias="_id")
    analysis_date: datetime
    owner: str
    causatives: Optional[List[str]] = None
    collaborators: List[str] = []
    individuals: List[Individual]
    genome_build: Annotated[str, BeforeValidator(lambda x: str(x))] = GENOME_BUILD_37
    panels: Optional[List[Panel]] = None
    rank_model_version: Optional[str] = None
    sv_rank_model_version: Optional[str] = None
    rank_score_threshold: int = 5
    phenotype_terms: Optional[List[Phenotype]] = None
    phenotype_groups: Optional[List[Phenotype]] = None
    diagnosis_phenotypes: Optional[List[DiagnosisPhenotypes]] = None
    diagnosis_genes: Optional[List[int]] = None


class Genotype(BaseModel):
    genotype_call: str
    allele_depths: List[int]
    read_depth: int
    genotype_quality: int
    sample_id: str


class Variant(BaseModel):
    document_id: str
    variant_id: str
    chromosome: str
    position: int
    dbsnp_id: Optional[str] = None
    reference: str
    alternative: str
    quality: Optional[float] = None
    filters: Optional[List[str]] = None
    end: int
    rank_score: int
    category: str
    sub_category: str
    genes: Optional[List[Gene]] = None
    samples: List[Genotype]



from sqlalchemy.orm import Query
from cg.store import Store
from cg.store.models import Invoice
from tests.store_helpers import StoreHelpers
from cg.store.filters.status_invoice_filters import (
    filter_invoices_by_invoice_id,
    filter_invoices_invoiced,
    filter_invoices_not_invoiced,
)
from tests.store.conftest import StoreConstants


def test_filter_get_invoices_by_invoice_id(
    store_with_an_invoice_with_and_without_attributes: Store,
    entry_id=StoreConstants.INVOICE_ID_INVOICE_WITH_ATTRIBUTES.value,
):
    """Test to get invoice by invoice id."""

    # GIVEN an Store with two invoices

    # WHEN getting invoice by invoice id
    invoice: Query = filter_invoices_by_invoice_id(
        invoices=store_with_an_invoice_with_and_without_attributes._get_query(table=Invoice),
        entry_id=entry_id,
    )

    # THEN assert that the invoice is a Query
    assert isinstance(invoice, Query)

    # THEN assert that the invoice is returned
    assert invoice.all() and len(invoice.all()) == 1


def test_filter_get_invoices_invoiced(store_with_an_invoice_with_and_without_attributes: Store):
    """Test to get invoice by invoice id."""

    # GIVEN an Store with two invoices of which one is invoiced

    # WHEN getting invoice by invoice id
    invoice: Query = filter_invoices_invoiced(
        invoices=store_with_an_invoice_with_and_without_attributes._get_query(table=Invoice)
    )

    # THEN assert that the invoice is a query
    assert isinstance(invoice, Query)

    # THEN assert that the invoice is returned
    assert invoice.all() and len(invoice.all()) == 1


def test_filter_get_invoices_not_invoiced(
    store_with_an_invoice_with_and_without_attributes: Store, helpers: StoreHelpers
):
    """Test to get invoice by invoice id."""

    # GIVEN a Store with two invoices of which one is invoiced

    # WHEN getting invoice by invoice id
    invoice: Query = filter_invoices_not_invoiced(
        invoices=store_with_an_invoice_with_and_without_attributes._get_query(table=Invoice)
    )

    # THEN assert that the invoice is a query
    assert isinstance(invoice, Query)

    # THEN assert that the invoice is returned
    assert invoice.all() and len(invoice.all()) == 1



"""Tests for the housekeeper storage functions of the demultiplexing post post-processing module."""

from pathlib import Path
from typing import List, Set

import mock
from housekeeper.store.models import File
from mock import MagicMock, call

from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.constants.housekeeper_tags import SequencingFileTag
from cg.meta.demultiplex.demux_post_processing import DemuxPostProcessingAPI
from cg.meta.demultiplex.housekeeper_storage_functions import (
    add_bundle_and_version_if_non_existent,
    add_demux_logs_to_housekeeper,
    add_sample_fastq_files_to_housekeeper,
    add_sample_sheet_path_to_housekeeper,
    add_tags_if_non_existent,
    get_sample_sheets_from_latest_version,
    store_fastq_path_in_housekeeper,
)
from cg.models.cg_config import CGConfig
from cg.models.demultiplex.flow_cell import FlowCellDirectoryData
from cg.store import Store


def test_add_bundle_and_version_if_non_existent(demultiplex_context: CGConfig):
    # GIVEN a DemuxPostProcessing API
    demux_post_processing_api = DemuxPostProcessingAPI(demultiplex_context)

    demux_post_processing_api.hk_api.bundle = MagicMock(return_value=None)
    demux_post_processing_api.hk_api.create_new_bundle_and_version = MagicMock()

    # WHEN adding a bundle and version which does not exist
    flow_cell_name: str = "flow_cell_name"
    add_bundle_and_version_if_non_existent(
        bundle_name=flow_cell_name, hk_api=demux_post_processing_api.hk_api
    )

    # THEN that the expected methods were called with the expected arguments
    demux_post_processing_api.hk_api.bundle.assert_called_once_with(name=flow_cell_name)
    demux_post_processing_api.hk_api.create_new_bundle_and_version.assert_called_once_with(
        name=flow_cell_name
    )


def test_add_bundle_and_version_if_already_exists(demultiplex_context: CGConfig):
    # GIVEN a DemuxPostProcessing API
    demux_post_processing_api = DemuxPostProcessingAPI(demultiplex_context)

    mock_bundle = MagicMock()
    demux_post_processing_api.hk_api.bundle = MagicMock(return_value=mock_bundle)
    demux_post_processing_api.hk_api.create_new_bundle_and_version = MagicMock()

    # WHEN adding a bundle and version which already exists
    flow_cell_name: str = "flow_cell_name"
    add_bundle_and_version_if_non_existent(
        bundle_name=flow_cell_name, hk_api=demux_post_processing_api.hk_api
    )
    # THEN the bundle was retrieved
    demux_post_processing_api.hk_api.bundle.assert_called_once_with(name=flow_cell_name)

    # THEN a new bundle and version was not created
    demux_post_processing_api.hk_api.create_new_bundle_and_version.assert_not_called()


def test_add_tags_if_non_existent(demultiplex_context: CGConfig):
    # GIVEN a DemuxPostProcessing API
    demux_post_processing_api = DemuxPostProcessingAPI(demultiplex_context)

    # GIVEN that the tags do not exist
    demux_post_processing_api.hk_api.get_tag = MagicMock(return_value=None)
    demux_post_processing_api.hk_api.add_tag = MagicMock()

    # WHEN adding new tags
    tag_names = ["tag1", "tag2"]
    add_tags_if_non_existent(tag_names=tag_names, hk_api=demux_post_processing_api.hk_api)

    # THEN the expected housekeeper API methods were called to create the tags
    demux_post_processing_api.hk_api.get_tag.assert_has_calls(
        [call(name="tag1"), call(name="tag2")]
    )
    demux_post_processing_api.hk_api.add_tag.assert_has_calls(
        [call(name="tag1"), call(name="tag2")]
    )


def test_add_tags_if_all_exist(demultiplex_context: CGConfig):
    # Given a DemuxPostProcessing API
    demux_post_processing_api = DemuxPostProcessingAPI(demultiplex_context)

    # Mock the methods in the housekeeper API
    demux_post_processing_api.hk_api.get_tag = MagicMock(return_value=MagicMock())
    demux_post_processing_api.hk_api.add_tag = MagicMock()

    # Call the add_tags_if_non_existent method with two tag names
    tag_names = ["tag1", "tag2"]
    add_tags_if_non_existent(tag_names=tag_names, hk_api=demux_post_processing_api.hk_api)

    # Assert that the expected methods were called with the expected arguments
    demux_post_processing_api.hk_api.get_tag.assert_has_calls(
        [call(name="tag1"), call(name="tag2")]
    )
    demux_post_processing_api.hk_api.add_tag.assert_not_called()


def test_add_fastq_files_without_sample_id(
    demultiplex_context: CGConfig, bcl_convert_flow_cell: FlowCellDirectoryData
):
    # GIVEN a DemuxPostProcessing API
    demux_post_processing_api = DemuxPostProcessingAPI(demultiplex_context)

    demux_post_processing_api.add_file_to_bundle_if_non_existent = MagicMock()

    # GIVEN that the sample sheet exists in housekeeper and the path is in the flow cell
    add_sample_sheet_path_to_housekeeper(
        flow_cell_directory=bcl_convert_flow_cell.path,
        flow_cell_name=bcl_convert_flow_cell.id,
        hk_api=demultiplex_context.housekeeper_api,
    )
    sample_sheet_path: Path = Path(
        get_sample_sheets_from_latest_version(
            flow_cell_id=bcl_convert_flow_cell.id, hk_api=demultiplex_context.housekeeper_api
        )[0].full_path
    )
    bcl_convert_flow_cell.set_sample_sheet_path_hk(hk_path=sample_sheet_path)

    # WHEN add_fastq_files is called

    add_sample_fastq_files_to_housekeeper(
        flow_cell=bcl_convert_flow_cell,
        hk_api=demux_post_processing_api.hk_api,
        store=demux_post_processing_api.status_db,
    )

    # THEN add_file_if_non_existent was not called
    demux_post_processing_api.add_file_to_bundle_if_non_existent.assert_not_called()


def test_add_existing_sample_sheet(
    demultiplex_context: CGConfig,
    bcl_convert_flow_cell: FlowCellDirectoryData,
    tmp_flow_cells_directory: Path,
):
    # GIVEN a DemuxPostProcessing API
    demux_post_processing_api = DemuxPostProcessingAPI(demultiplex_context)

    # GIVEN a flow cell directory and name
    flow_cell_directory = Path(tmp_flow_cells_directory, bcl_convert_flow_cell.full_name)

    # GIVEN that a flow cell bundle exists in Housekeeper
    add_bundle_and_version_if_non_existent(
        bundle_name=bcl_convert_flow_cell.id, hk_api=demux_post_processing_api.hk_api
    )

    # WHEN a sample sheet is added
    add_sample_sheet_path_to_housekeeper(
        flow_cell_directory=flow_cell_directory,
        flow_cell_name=bcl_convert_flow_cell.id,
        hk_api=demux_post_processing_api.hk_api,
    )

    # THEN a sample sheet file was added to the bundle
    expected_tag_names = [SequencingFileTag.SAMPLE_SHEET, bcl_convert_flow_cell.id]

    files = demux_post_processing_api.hk_api.get_files(
        bundle=bcl_convert_flow_cell.id, tags=expected_tag_names
    ).all()
    assert len(files) == 1


def test_add_demux_logs_to_housekeeper(
    demultiplex_context: CGConfig, bcl_convert_flow_cell: FlowCellDirectoryData
):
    # GIVEN a DemuxPostProcessing API
    demux_post_processing_api = DemuxPostProcessingAPI(demultiplex_context)

    # GIVEN a bundle and flow cell version exists in housekeeper
    add_bundle_and_version_if_non_existent(
        bundle_name=bcl_convert_flow_cell.id, hk_api=demux_post_processing_api.hk_api
    )

    # GIVEN a demux log in the run directory
    demux_log_file_paths: List[Path] = [
        Path(
            demux_post_processing_api.flow_cells_dir,
            f"{bcl_convert_flow_cell.full_name}",
            f"{bcl_convert_flow_cell.id}_demultiplex.stdout",
        ),
        Path(
            demux_post_processing_api.flow_cells_dir,
            f"{bcl_convert_flow_cell.full_name}",
            f"{bcl_convert_flow_cell.id}_demultiplex.stderr",
        ),
    ]
    for file_path in demux_log_file_paths:
        file_path.parent.mkdir(parents=True, exist_ok=True)
        file_path.touch()

    # WHEN adding the demux logs to housekeeper
    add_demux_logs_to_housekeeper(
        flow_cell=bcl_convert_flow_cell,
        flow_cell_run_dir=demux_post_processing_api.flow_cells_dir,
        hk_api=demux_post_processing_api.hk_api,
    )

    # THEN the demux log was added to housekeeper
    files = demux_post_processing_api.hk_api.get_files(
        tags=[SequencingFileTag.DEMUX_LOG],
        bundle=bcl_convert_flow_cell.id,
    ).all()

    expected_file_names: List[str] = []
    for file_path in demux_log_file_paths:
        expected_file_names.append(file_path.name.split("/")[-1])

    # THEN the demux logs were added to housekeeper with the correct names
    assert len(files) == 2
    for file in files:
        assert file.path.split("/")[-1] in expected_file_names


def test_store_fastq_path_in_housekeeper_correct_tags(
    populated_housekeeper_api: HousekeeperAPI,
    empty_fastq_file_path: Path,
    novaseq6000_flow_cell: FlowCellDirectoryData,
):
    """Test that a fastq file is stored in Housekeeper with the correct tags."""
    sample_id: str = "sample_internal_id"
    # GIVEN a fastq file that has not been added to Housekeeper
    assert not populated_housekeeper_api.files(path=empty_fastq_file_path.as_posix()).first()

    # WHEN adding the fastq file to housekeeper
    store_fastq_path_in_housekeeper(
        sample_internal_id=sample_id,
        sample_fastq_path=empty_fastq_file_path,
        flow_cell_id=novaseq6000_flow_cell.id,
        hk_api=populated_housekeeper_api,
    )

    # THEN the file was added to Housekeeper with the correct tags
    file: File = populated_housekeeper_api.get_files(bundle=sample_id).first()
    expected_tags: Set[str] = {SequencingFileTag.FASTQ.value, novaseq6000_flow_cell.id, sample_id}
    assert set([tag.name for tag in file.tags]) == expected_tags
